from zenlog import log


def convert_types(obj, only_nx=True):
    from numpy import dtype, ndarray, array
    from nexusformat.nexus import NXattr
    py_data_type = type(obj)
    np_data_type = dtype(py_data_type)
    if np_data_type.name == 'object':
        if isinstance(obj, ndarray):
            vl = obj.tolist()
            if len(vl):
                el = vl[0]
                for ii in range(1, len(obj.shape)):
                    el = el[0]
                tp = dtype(type(el)).name
            else:
                tp = obj.dtype.name
        elif obj is None:
            (tp, vl) = ('string', 'None')
        elif isinstance(obj, NXattr):
            val = obj.nxdata
            if isinstance(val, ndarray):
                val = val.tolist()
            if obj.dtype == 'object':
                (tp, vl) = (dtype(type(obj.nxdata)).name, val)
                # If still 'object', this will throw an error below
                if tp == 'object' and isinstance(val, list):
                    tp = dtype(type(val[0])).name
            else:
                (tp, vl) = (obj.dtype, val)
        elif not only_nx and hasattr(obj, 'to_json_dict'):
            # Shoe-horn in an object-defined dictionary:
            tp, vl = None, obj.to_json_dict()
        elif isinstance(obj, list):
            return convert_types(array(obj))
        else:
            raise RuntimeError(f'unrecognised type {py_data_type} / {np_data_type} for {repr(obj)}')
    else:
        (tp, vl) = (np_data_type.name, obj)
    if tp == 'str':
        tp = 'string'
    elif tp == 'float64':
        tp = 'double'
    elif tp == 'object':
        raise RuntimeError(f'Internal logical error attempting to convert {obj} of type {type(obj)}')
    elif tp == 'int':
        tp = 'int64'
    elif tp == 'float':
        tp = 'double'
    return tp, vl


class Writer:
    """
    Writes out files in various formats from a NeXus structure with instrument information
    """

    def __init__(self, nx_obj):
        from nexusformat.nexus import NXsample, NXentry, NXroot
        self.root_name = 'entry'
        self.instrument = None
        self.data = None
        self.sample = NXsample()
        self.nx_obj = None
        if nx_obj.nxclass == 'NXinstrument':
            self.instrument = nx_obj
            nx_entry = NXentry()
            nx_entry['instrument'] = self.instrument
        else:
            if nx_obj.nxclass == 'NXroot':
                entries = dir(nx_obj)
                if len(entries) > 1:
                    log.warn('More than one entry in nx_obj, using only first entry')
                self.root_name = entries[0]
                self.nx_obj = nx_obj
                nx_entry = nx_obj[self.root_name]
            elif nx_obj.nxclass == 'NXentry':
                nx_entry = nx_obj
            else:
                raise RuntimeError('Input must be an NXroot, NXentry or NXinstrument')
            for x in ('instrument', 'data', 'sample'):
                if hasattr(nx_entry, x):
                    setattr(self, x, nx_entry[x])

        if self.instrument is None:
            raise RuntimeError('No instrument found')

        if self.nx_obj is None:
            self.nx_obj = NXroot()
            self.nx_obj[self.root_name] = nx_entry

    def to_json(self, filename, indent=4, only_nx=True, absolute_depends_on=False):
        """Convert a NeXus object to a JSON-compatible dictionary, then write that to file

        Parameters:
            filename: str - where to write the JSON string, '.json' will be appended if not present
            indent: int - nested JSON indentation depth, default 4
            only_nx: bool - whether non-NeXus objects found in the tree raise an error, default=True
            absolute_depends_on: bool - if True expand `depends_on` clauses to absolute paths
        """
        from json import dumps
        if not filename.endswith('.json'):
            filename = f'{filename}.json'
        to_write = dumps(self.to_nexus_structure(only_nx, absolute_depends_on), indent=indent)
        with open(filename, 'w') as file:
            file.write(to_write)

    def to_nexus_structure(self, only_nx=True, absolute_depends_on=False):
        """Convert a NeXus object to a JSON-compatible 'NeXus-Structure' dictionary as used by the ESS file-writer"""
        json_dict = self._to_json_dict(self.nx_obj, only_nx=only_nx, absolute_depends_on=absolute_depends_on)
        return dict(children=json_dict)

    def _to_json_dict(self, top_obj, only_nx=True, absolute_depends_on=False):
        """Recursive transversal of NXobject tree conversion to JSON-compatible dict"""
        # Note to Greg, depends_on can be data or attribute
        children = []
        for name, obj in top_obj.items():
            if hasattr(obj, 'nxclass'):
                attrs = []
                if absolute_depends_on and 'depends_on' == name and not obj.nxdata.startswith('/'):
                    obj.nxdata = _to_absolute(top_obj.nxpath, obj.nxdata)
                if obj.nxclass == 'NXfield':
                    typ, val = convert_types(obj.nxdata, only_nx)
                    # typ is None if obj.nxdata is a NotNXdict (such that val _is_ the contained dict)
                    entry = dict(module='dataset', config=dict(name=name, values=val, type=typ)) if typ else val
                else:
                    entry = dict(name=name, type='group')
                    attrs = [dict(name='NX_class', dtype='string', values=obj.nxclass)]
                    if len(list(obj)):
                        entry['children'] = self._to_json_dict(obj, only_nx=only_nx, absolute_depends_on=absolute_depends_on)
                for n in obj.attrs:
                    typ, val = convert_types(obj.attrs[n], only_nx)
                # FIXME accessing an attribute value via the dict values gives
                #       a NXattr object *not* the underlying value!?
                # for n, v in obj.attrs.items():
                #     typ, val = convert_types(v, only_nx)
                    if absolute_depends_on and n == 'depends_on' and '/' != val[0]:
                        val = _to_absolute(top_obj.nxpath, val)
                    attrs.append(dict(name=n, dtype=typ, values=val) if typ else val)
                if len(attrs):
                    entry['attributes'] = attrs
            elif not only_nx and hasattr(obj, 'to_json_dict'):
                # This branch is unreachable because any Python object added to a NXobject gets wrapped in NXfield
                entry = obj.to_json_dict()
            else:
                raise RuntimeError(f'Unrecognized object key {name}')
            children.append(entry)
        return children


def _to_absolute(parent: str, path: str):
    if '.' == path:
        return '.'
    return f'{parent}/{path}'




def get_bifrost_instr():
    from mccode_antlr.io.json import load_json
    from pathlib import Path
    file = Path(__file__).parent / 'bifrost.instr.json'
    return load_json(file)

def test_bifrost_nexus_structure():
    # from pathlib import Path
    # from json import dump
    from moreniius import MorEniius
    instr = get_bifrost_instr()
    me = MorEniius.from_mccode(instr, origin='sample_origin', only_nx=False, absolute_depends_on=False)
    ns = me.to_nexus_structure()

    # jq path to pulse_shaping_chopper_1's rotation_speed
    #.children[0].children[0].children[12].children[2]
    obj = ns['children'][0]['children'][0]['children'][12]['children'][2]


    assert 'type' in obj
    assert obj['type'] == 'group'
    children = obj['children']
    for child in children:
        assert 'module' in child
        assert 'config' in child
        assert child['module'] == 'link'
        assert all(x in child['config'] for x in ('name', 'source'))
        assert child['config']['name'] == child['config']['source'].split('/')[-1]

    # The channel-5 arm is oriented 0-degrees rotated from the tank rotation angle
    # and _should_ be a NXccordinate_system directly dependent on that NXlog'ed value
    # jless finds it at .children[0].children[0].children[255]
    arm = ns['children'][0]['children'][0]['children'][255]
    assert 'type' in arm
    assert arm['type'] == 'group'
    assert 'name' in arm
    assert arm['name'] == 'channel_5_arm'

    assert 'attributes' in arm
    class_name = None
    for attribute in arm['attributes']:
        if (n := attribute.get('name')) is not None and n == 'NX_class':
            class_name = attribute['values']
    assert class_name == 'NXcoordinate_system'

    assert 'children' in arm
    for child in arm['children']:
        if (m := child.get('module')) is not None and m == 'dataset':
            assert 'config' in child
            if (n := child['config'].get('name')) is not None and n == 'depends_on':
                # Check for the actual named group?
                assert child['config']['values'].startswith('/')



from __future__ import annotations

from zenlog import log
from dataclasses import dataclass, field
from mccode_antlr.instr import Orient, Parts, Part
from nexusformat.nexus import NXfield
from .instr import NXInstr

@dataclass
class NXPart:
    instr: NXInstr
    o: Part

    def expr2nx(self, expr):
        return self.instr.expr2nx(expr)

    def make_nx(self, nx_class, *args, **kwargs):
        return self.instr.make_nx(nx_class, *args, **kwargs)

    def make_translation(self, norm, vec, dep):
        # if `norm` is a link or NXlog, we should make a group not an NXfield
        return self.make_nx(NXfield, norm, vector=vec, depends_on=dep, transformation_type='translation', units='m')

    def translations(self, dep: str, name: str) -> list[tuple[str, NXfield]]:
        from mccode_antlr.instr import RotationPart
        from mccode_antlr.common import Expr, Value
        if isinstance(self.o, RotationPart):
            raise RuntimeError('Part is a rotation!')
        pos = self.o.position()
        if any(isinstance(c, (Expr, Value)) for c in (pos.x, pos.y, pos.z)):
            translations = []
            for n, c, v in (('x', pos.x, [1, 0, 0]), ('y', pos.y, [0, 1, 0]), ('z', pos.z, [0, 0, 1])):
                if c != Expr.parse('0'):
                    next_name = f'{name}_{n}'
                    translations.append((next_name, self.make_translation(c, v, dep)))
                    dep = next_name
            return translations
        # vector is all constants, hopefully
        norm = pos.length()
        vec = pos if norm.is_zero else pos/norm
        return [(name, self.make_translation(norm, vec, dep))]

    def rotation(self, dep: str) -> NXfield:
        from mccode_antlr.instr import TranslationPart
        if isinstance(self.o, TranslationPart):
            raise RuntimeError('Part is a translation')
        try:
            axis, angle, angle_unit = self.o.rotation_axis_angle
        except RuntimeError as error:
            log.error(f'Failed to get rotation axis and angle: {error}')
            print(repr(self.o))
            raise NotImplementedError()

        # handle the case where angle is not a constant?
        return self.make_nx(NXfield, angle, vector=axis, depends_on=dep, transformation_type='rotation', units=angle_unit)

    def transformations(self, name: str, dep: str | None = None) -> list[tuple[str, NXfield]]:
        if self.o.is_translation and self.o.is_rotation:
            ops = self.translations(dep, name)
            rot = self.rotation(ops[-1][0])
            return [*ops, (f'{name}_r', rot)]
        elif self.o.is_translation:
            return self.translations(dep, name)
        elif self.o.is_rotation:
            return [(name, self.rotation(dep))]
        else:
            return []


@dataclass
class NXParts:
    instr: NXInstr
    position: Parts
    rotation: Parts

    def _transformations(self, name: str, dep: str, typ: str, stack):
        nx_transformations = []
        for i, op in enumerate(stack):
            parts = NXPart(self.instr, op).transformations(f'{name}_{typ}{i}', dep)
            nx_transformations.extend(parts)
            if len(parts) and len(parts[-1]):
                dep = parts[-1][0]
        return nx_transformations

    def position_transformations(self, name: str, dep: str | None = None ) -> list[tuple[str, NXfield]]:
        dep = dep or '.'
        return self._transformations(name, dep, 't', self.position.stack())

    def rotation_transformations(self, name: str, dep: str | None = None) -> list[tuple[str, NXfield]]:
        dep = dep or '.'
        return self._transformations(name, dep, 'r', self.rotation.stack())

    def transformations(self, name: str, dep: str | None = None) -> list[tuple[str, NXfield]]:
        parts = self.position_transformations(name, dep=dep)
        # If there were any positioning transformations, we need to update
        # the dependency chained name, otherwise it should stay the same
        dep = parts[-1][0] if len(parts) and len(parts[-1]) else dep
        parts += self.rotation_transformations(name, dep=dep)
        if dep is not None and len(parts) == 0:
            # We were told that we should have a relative transformation
            # but the position and rotation are all zeros!
            return [(dep, NXfield())]
        return parts


@dataclass
class NXOrient:
    instr: NXInstr
    do: Orient
    nx_parts: NXParts | None = None

    def __post_init__(self):
        # collapse all possible chained orientation information
        # But keep the rotations and translations separate
        pos, rot = self.do.position_parts(), self.do.rotation_parts()
        # make an ordered list of the requisite NXfield entries
        self.nx_parts = NXParts(self.instr, pos, rot)

    def transformations(self, name: str) -> dict[str, NXfield]:
        # make an ordered list of the requisite NXfield entries, and turn it into a dict
        return {k: v for k, v in self.nx_parts.transformations(name)}

    def position_transformations(self, name: str, dep: str | None = None) -> list[tuple[str, NXfield]]:
        return self.nx_parts.position_transformations(name, dep)

    def rotation_transformations(self, name: str, dep: str | None = None) -> list[tuple[str, NXfield]]:
        return self.nx_parts.rotation_transformations(name, dep)





from zenlog import log
from .mccode.instance import COMPONENT_TYPE_NAME_TO_NEXUS, register_translator
from .mccode.comp import monitor_translator

log.debug('Extending moreniius.mccode.NXInstance translators')

# Tell the converter about missing component(s)
COMPONENT_TYPE_NAME_TO_NEXUS['ESS_butterfly'] = 'NXmoderator'


# BIFROST has 45 triplet detector modules. Each has its own position and orientation, and defines the positions
# and orientations of its cylindrical pixels in its own coordinate system.
# This module-level dictionary is used to dynamically populate their information, to be used when constructing the
# full detector during the readout-construction.
BIFROST_DETECTOR_MODULES = {}

BIFROST_DETECTOR_TOPIC = 'bifrost_detector'


def readout_translator(instance):
    """BIFROST specific Readout Master, should be deprecated in favour of ReadoutCAEN"""
    from numpy import einsum, ndarray, array
    from nexusformat.nexus import NXdetector, NXcylindrical_geometry
    from .utils import ev44_event_data_group
    stream = ev44_event_data_group(source='caen', topic=BIFROST_DETECTOR_TOPIC)

    readout_pos, readout_rot = instance.obj.orientation.position_parts(), instance.obj.orientation.rotation_parts()

    counts = [geometry.detector_number.size for _, _, geometry in BIFROST_DETECTOR_MODULES.values()]
    total = sum(counts)
    points = [geometry.vertices.shape[0] for _, _, geometry in BIFROST_DETECTOR_MODULES.values()]
    total_points = sum(points)
    all_cylinders = ndarray((total, 3), dtype='int32')
    all_vertices = ndarray((total_points, 3), dtype='float32')
    detector_number = ndarray((total,), dtype='int32')

    offset = 0

    for module_pos, module_rot, geometry in BIFROST_DETECTOR_MODULES.values():
        # Find the vector and rotation matrix linking the two coordinate systems
        #
        # - The geometry is defined in the module coordinate system, where any point is p' = (x', y', z').
        # - There is a rotation matrix rot1 that relates a vector in _a_ global coordinate system to the module
        #   Such that
        #       x' = rot1 x_global
        #   Or
        #       x_global = rot1.inv() x'
        # - Similarly, there is a rotation matrix rot2 that relates a vector in the same global coordinate system to
        #   the readout's coordinate system.
        #       x = rot2 x_global
        # - In this global coordinate system, the module position and readout positions have a vector connecting them
        #       v = p_module - p_readout
        #
        # - So, to express the geometry in the readout's coordinate system, we convert p' to the global coordinate
        #   system, add the difference vector between the two coordinate systems, then convert the result to the
        #   readout coordinate system:
        #       p = rot2 (v + rot1.inv() p')
        #
        v = instance.instr.expr2nx(tuple((module_pos - readout_pos).position()))  # this is a mccode-antlr Vector
        rot1 = array(instance.instr.expr2nx(tuple(module_rot.inverse().rotation()))).reshape((3, 3))
        rot2 = array(instance.instr.expr2nx(tuple(readout_rot.rotation()))).reshape((3, 3))
        #
        vertices = einsum('ij,kj->ki', rot1, geometry.vertices)
        vertices = vertices + v
        vertices = einsum('ij,kj->ki', rot2, vertices)
        #
        cylinders = offset + geometry.cylinders
        all_vertices[offset:(offset + vertices.shape[0])] = vertices
        all_cylinders[geometry.detector_number - 1, :] = cylinders
        detector_number[geometry.detector_number - 1] = geometry.detector_number
        offset += vertices.shape[0]

    geometry = NXcylindrical_geometry(vertices=all_vertices, cylinders=all_cylinders, detector_number=detector_number)

    return NXdetector(data=stream, geometry=geometry)


def monochromator_rowland_translator(nxinstance):
    from nexusformat.nexus import NXcrystal
    from .nxoff import NXoff
    h = nxinstance.nx_parameter('yheight') / 2
    w = nxinstance.nx_parameter('zwidth') / 2
    count = nxinstance.nx_parameter('NH')
    gap = nxinstance.nx_parameter('gap')

    crystal = NXcrystal()
    crystal['usage'] = 'Bragg'
    crystal['d_spacing'] = nxinstance.nx_parameter('DM')
    crystal['d_spacing'].units = 'angstrom'
    crystal['segment_width'] = w * 2
    crystal['segment_width'].units = 'm'
    crystal['segment_height'] = h * 2
    crystal['segment_height'].units = 'm'
    crystal['segment_gap'] = gap
    crystal['segment_gap'].units = 'm'
    crystal['segment_columns'] = count
    crystal['segment_rows'] = nxinstance.nx_parameter('NV')
    mosaic_h = nxinstance.nx_parameter('mosaich')
    mosaic_v = nxinstance.nx_parameter('mosaicv')
    mosaic = nxinstance.nx_parameter('mosaic')
    crystal['mosaic_horizontal'] = mosaic_h if mosaic_h else mosaic
    crystal['mosaic_horizontal'].units = 'arcminutes'
    crystal['mosaic_vertical'] = mosaic_v if mosaic_v else mosaic
    crystal['mosaic_vertical'].units = 'arcminutes'
    # ... could add curvature by copying calculation ... but we don't have source or sink distances here.

    # nexus-constructor does not use any of the segment information in its drawing -- so we need to specify OFF
    # the danger of course is that without the distances we *can't* get the rowland geometry right and someone might
    # assume that the OFF is more-valid than the McStas data :/
    vertices = []
    faces = []
    for i in range(count):
        x0 = (i - count // 2) * (2 * w + gap)
        # "For an unrotated monochromator component, the crystal surface lies in the Y-Z plane" -- McDoc
        vertices.extend([[0, -h, x0 - w], [0, -h, x0 + w], [0, h, x0 + w], [0, h, x0 - w]])
        faces.append([4*i, 4*i+1, 4*i+2, 4*i+3])

    geometry = NXoff(vertices, faces)
    crystal['geometry'] = geometry.to_nexus()

    return crystal


def bifrost_pixel_regex_20230703():
    from re import compile
    # This is highly specialized to BIFROST_nxs as of 2023-07-03
    # where we always have "WHEN (# == secondary_cassette && # == analyzer)
    w = r"WHEN \((?P<cassette>[0-9]+)\s*==\s*secondary_cassette\s*&&\s*(?P<analyzer>[0-9]+)\s*==\s*analyzer\)"
    r = compile(w)

    def icd_pixel(resolution, arc, triplet, tube, position):
        # position takes values between 0 and resolution - 1
        return 27 * resolution * arc + 9 * resolution * tube + resolution * triplet + position + 1

    return icd_pixel, r


def bifrost_pixel_regex_20230911():
    from re import compile
    # This is highly specialized to BIFROST_nxs as of 2023-07-03
    # where we always have "WHEN (# == secondary_cassette && # == analyzer)
    w = r"(?P<cassette>[0-9]+)\s*==\s*secondary_cassette\s*&&\s*(?P<analyzer>[0-9]+)\s*==\s*analyzer"
    r = compile(w)

    def icd_pixel(resolution, arc, triplet, tube, position):
        # position takes values between 0 and resolution - 1
        return 27 * resolution * arc + 9 * resolution * tube + resolution * triplet + position + 1

    return icd_pixel, r


def bifrost_source_20230704(arc, triplet):
    return f"arc={arc};triplet={triplet}"


def detector_tubes_offsets_and_one_cylinder(self):
    import numpy as np
    from nexusformat.nexus import NXdetector, NXfield, NXcylindrical_geometry
    from .utils import ev44_event_data_group
    # parameters for NXdetector, to be filled-in
    pars = {}
    pixel_fun, wre = bifrost_pixel_regex_20230911()

    cassette = 1
    analyzer = 1
    if wre.match(str(self.obj.when)):
        m = wre.match(str(self.obj.when))
        cassette = int(m.group('cassette'))
        analyzer = int(m.group('analyzer'))

    # cassette in (1, 9), analyzer in (1, 5):

    # i is the 'slow' direction, j is the 'fast' direction
    # --> i between tubes, j along tubes
    ni = self.nx_parameter('N') # corresponds to 'width' and McStas 'x' axis
    nj = self.nx_parameter('no')  # corresponds to 'height' and McStas 'y' axis
    width, height, radius = [self.nx_parameter(n) for n in ('width', 'height', 'radius')]

    halfi = (width - 2 * radius) / 2
    di = np.linspace(-halfi, halfi, ni)

    half_pixel = height / (nj + 1) / 2  # half the size of one pixel along the tube
    dj = -np.linspace(-height / 2 + half_pixel, height / 2 - half_pixel, nj)
    Dj, Di = np.meshgrid(dj, di)

    arc, triplet = analyzer - 1, cassette - 1  # naming from ICD 01 v6 indexing of triplets

    diameter = f'2 * {radius}' if isinstance(radius, str) else 2 * radius

    detector_number = [[pixel_fun(nj, arc, triplet, tube, position) for position in range(nj)] for tube in range(ni)]

    pars['detector_number'] = np.array(detector_number).astype('int32')
    pars['data'] = ev44_event_data_group(bifrost_source_20230704(arc, triplet), BIFROST_DETECTOR_TOPIC)
    pars['x_pixel_offset'] = NXfield(Di, units='m')
    pars['y_pixel_offset'] = NXfield(Dj, units='m')
    pars['x_pixel_size'] = NXfield(diameter, units='m')
    pars['y_pixel_size'] = NXfield(height / nj, units='m')
    pars['diameter'] = NXfield(diameter, units='m')
    pars['type'] = f'{ni} He3 tubes in series' if self.nx_parameter('wires_in_series', True) else f'{ni} He3 tubes'

    # use NXcylindrical_geometry to define the detectors, which requires:
    #   vertices - (i, 3) -- points relative to the detector position defining each cylinder in the detector
    #   cylinders - (j, 3) -- indexes of the vertices, to define a cylinder by its face-center, face-edge, and
    #                         opposite face center:
    #             |---------------|
    #             |               |
    #           0 +               + 2
    #             |               |
    #           1 |---------------|
    #   detector_number: (k,) -- maps the cylinders in cylinder by index with a detector id
    #
    # We're allowed to specify a single cylinder then le the x/y/z_offset position that pixel repeatedly:
    vertices = NXfield([[0, -half_pixel, 0], [radius, -half_pixel, 0], [0, half_pixel, 0]], units='m')
    cylinders = [[0, 1, 2]]
    geometry = NXcylindrical_geometry(vertices=vertices, cylinders=cylinders)

    return NXdetector(**pars, geometry=geometry)

#
# def detector_tubes_only_cylinder(self):
#     """This results in only the first cylinder being plotted by Nexus constructor"""
#     import numpy as np
#     from nexusformat.nexus import NXdetector, NXfield, NXcylindrical_geometry
#     from .utils import ev44_event_data_group
#     # parameters for NXdetector, to be filled-in
#     pars = {}
#     pixel_fun, wre = bifrost_pixel_regex_20230911()
#
#     if wre.match(str(self.obj.when)):
#         m = wre.match(str(self.obj.when))
#         cassette = int(m.group('cassette'))
#         analyzer = int(m.group('analyzer'))
#     else:
#         cassette = 1
#         analyzer = 1
#     # cassette in (1, 9), analyzer in (1, 5):
#
#     # i is the 'slow' direction, j is the 'fast' direction
#     # --> i between tubes, j along tubes
#     ni = self.nx_parameter('N') # corresponds to 'width' and McStas 'x' axis
#     nj = self.nx_parameter('no')  # corresponds to 'height' and McStas 'y' axis
#     width, height, radius = [self.nx_parameter(n) for n in ('width', 'height', 'radius')]
#     halfi = (width - 2 * radius) / 2
#     di = np.linspace(-halfi, halfi, ni)
#     dj = np.linspace(-height / 2, height / 2, nj+1)
#
#     # use NXcylindrical_geometry to define the detectors, which requires:
#     #   vertices - (i, 3) -- points relative to the detector position defining each cylinder in the detector
#     #   cylinders - (j, 3) -- indexes of the vertices, to define a cylinder by its face-center, face-edge, and
#     #                         opposite face center:
#     #             |---------------|---------------|
#     #             |               |               |
#     #           0 +               + 2             + 4
#     #             |               |               |
#     #             |---------------|---------------|
#     #             1               3               5
#     #   detector_number: (k,) -- maps the cylinders in cylinder by index with a detector id
#
#     arc, triplet = analyzer - 1, cassette - 1  # naming from ICD 01 v6 indexing of triplets
#
#     vertices = NXfield([v for x in di for y in dj for v in [[x, y, 0], [x, y, radius]]], units='m')
#     cylinders = [[k, k+1, k+2] for k in [tube * 2 * (nj + 1) + 2 * j for tube in range(ni) for j in range(nj)]]
#     detector_number = [pixel_fun(nj, arc, triplet, tube, j) for tube in range(ni) for j in range(nj)]
#
#     pars['data'] = ev44_event_data_group(bifrost_source_20230704(arc, triplet), BIFROST_DETECTOR_TOPIC)
#     pars['type'] = f'{ni} He3 tubes in series' if self.nx_parameter('wires_in_series') else f'{ni} He3 tubes'
#
#     geometry = NXcylindrical_geometry(vertices=vertices, cylinders=cylinders, detector_number=detector_number)
#
#     return NXdetector(**pars, geometry=geometry)


def detector_tubes_only_cylinder(self):
    """This results in only the first cylinder being plotted by Nexus constructor"""
    import numpy as np
    from nexusformat.nexus import NXdetector, NXfield, NXcylindrical_geometry
    from .utils import ev44_event_data_group
    # parameters for NXdetector, to be filled-in
    pars = {}
    pixel_fun, wre = bifrost_pixel_regex_20230911()

    if wre.match(str(self.obj.when)):
        m = wre.match(str(self.obj.when))
        cassette = int(m.group('cassette'))
        analyzer = int(m.group('analyzer'))
    else:
        cassette = 1
        analyzer = 1
    # cassette in (1, 9), analyzer in (1, 5):

    # i is the 'slow' direction, j is the 'fast' direction
    # --> i between tubes, j along tubes
    ni = self.nx_parameter('N') # corresponds to 'width' and McStas 'x' axis
    nj = self.nx_parameter('no')  # corresponds to 'height' and McStas 'y' axis
    width, height, radius = [self.nx_parameter(n) for n in ('width', 'height', 'radius')]
    halfi = (width - 2 * radius) / 2
    # signs of di and dj verified by examining plots of detector positions
    di = np.linspace(-halfi, halfi, ni)
    dj = -np.linspace(-height / 2, height / 2, nj+1)

    # use NXcylindrical_geometry to define the detectors, which requires:
    #   vertices - (i, 3) -- points relative to the detector position defining each cylinder in the detector
    #   cylinders - (j, 3) -- indexes of the vertices, to define a cylinder by its face-center, face-edge, and
    #                         opposite face center:
    #             |---------------|---------------|
    #             |               |               |
    #           0 +               + 2             + 4
    #             |               |               |
    #             |---------------|---------------|
    #             1               3               5
    #   detector_number: (k,) -- maps the cylinders in cylinder by index with a detector id

    arc, triplet = analyzer - 1, cassette - 1  # naming from ICD 01 v6 indexing of triplets

    vertices = NXfield([v for x in di for y in dj for v in [[x, y, 0], [x, y, radius]]], units='m')
    cylinders = np.array([[k, k+1, k+2] for k in [tube * 2 * (nj + 1) + 2 * j for tube in range(ni) for j in range(nj)]]).astype('int32')
    detector_number = np.array([pixel_fun(nj, arc, triplet, tube, j) for tube in range(ni) for j in range(nj)]).astype('int32')

    pars['data'] = ev44_event_data_group(bifrost_source_20230704(arc, triplet), BIFROST_DETECTOR_TOPIC)
    pars['type'] = f'{ni} He3 tubes in series' if self.nx_parameter('wires_in_series', True) else f'{ni} He3 tubes'

    geometry = NXcylindrical_geometry(vertices=vertices, cylinders=cylinders, detector_number=detector_number)

    for md in self.obj.metadata:
        if md.mimetype == 'application/json' and md.name == 'nexus_structure_stream_data':
            from nexusformat.nexus import NXdata
            from moreniius.utils import NotNXdict
            from json import loads
            stream = loads(md.value)
            name = stream.get('module', 'metadata')
            pars[name] = NXdata(data=NotNXdict(stream))

    return NXdetector(**pars, geometry=geometry)


def bifrost_detector_collector(self):
    # nx_obj = detector_tubes_only_cylinder(self)  # each pixel is a cylinder
    nx_obj = detector_tubes_offsets_and_one_cylinder(self)  # all pixels share one cylinder
    # stash the object parts
    pos, rot = self.obj.orientation.position_parts(), self.obj.orientation.rotation_parts()
    BIFROST_DETECTOR_MODULES[str(self.obj.when)] = pos, rot, nx_obj.geometry
    # and return it
    return nx_obj


# def histogram_monitor(obj):
#     from nexusformat.nexus import NXmonitor
#     from .nxoff import NXoff
#     from .utils import ev44_event_data_group
#     width = obj.nx_parameter('xwidth')
#     height = obj.nx_parameter('yheight')
#     geometry = NXoff.from_wedge(l=0.005, w1=width, h1=height)
#
#     # parameters to be filled-in
#     pars = {}
#     # pars['data'] = ev44_event_data_group(bifrost_source_20230704(arc, triplet), BIFROST_DETECTOR_TOPIC)
#     # pars['type'] = f'{ni} He3 tubes in series' if self.nx_parameter('wires_in_series') else f'{ni} He3 tubes'
#     pars['geometry'] = geometry.to_nexus()
#     return NXmonitor(**pars)


# Patch-in the new methods
# register_translator('Readout', readout_translator)
register_translator('Monochromator_Rowland', monochromator_rowland_translator)
register_translator('Detector_tubes', bifrost_detector_collector)
register_translator('Detector_time_tubes', bifrost_detector_collector)
register_translator('Frame_monitor', monitor_translator)

log.debug('moreniius.mccode.NXInstance translators extended')



#!/usr/bin/env python
# coding: utf-8

from __future__ import annotations
from pathlib import Path
from mccode_antlr.instr import Instr


def load_instr(filepath: str | Path) -> Instr:
    """Loads an Instr object from a .instr file or a HDF5 file"""
    if not isinstance(filepath, Path):
        filepath = Path(filepath)
    if not filepath.exists() or not filepath.is_file():
        raise ValueError('The provided filepath does not exist or is not a file')

    if filepath.suffix == '.instr':
        from mccode_antlr.loader import load_mcstas_instr
        return load_mcstas_instr(filepath)
    elif filepath.suffix.lower() == '.json':
        from mccode_antlr.io.json import load_json
        return load_json(filepath)

    from mccode_antlr.io import load_hdf5
    return load_hdf5(filepath)


def to_nexus_structure(instr: Instr) -> dict:
    from moreniius import additions
    from moreniius import MorEniius
    nx = MorEniius.from_mccode(instr, origin='sample_stack', only_nx=False, absolute_depends_on=True)
    return nx.to_nexus_structure()


def convert():
    """Convert an Instr (HDF5|JSON) or .instr (text) file to an equivalent NeXus Structure JSON string"""
    import argparse
    parser = argparse.ArgumentParser(description='Convert an Instr (HDF5|JSON) or .instr (text) file to an equivalent NeXus Structure JSON string')
    parser.add_argument('filename', type=str, help='the file to convert')
    parser.add_argument('--format', type=str, default='json', help='the output format, currently only json')
    args = parser.parse_args()
    instr = load_instr(args.filename)
    structure = to_nexus_structure(instr)
    if args.format == 'json':
        from json import dumps
        print(dumps(structure))
    else:
        raise RuntimeError(f"Unknown format {args.format}")


if __name__ == '__main__':
    convert()
